import numpy as np
from .history import eg48_mean, phase_to_subsidy
from .timeline import Timeline

def get_new_difficulty_bits(zero_bits, timespan):
    '''
    Compute the new value of 32 + log2(difficulty)
    '''

    blocks_per_epoch = 2016
    seconds_in_epoch = blocks_per_epoch * 600.0

    # Only 2015 of the 2016 blocks actually contribute to the difficulty
    # calculation -- see https://www.bitrawr.com/difficulty-estimator --
    # so we apply this adjustment factor:
    adjustment_factor = blocks_per_epoch / (blocks_per_epoch - 1.0)

    adjustment = np.log(seconds_in_epoch * adjustment_factor / timespan) / np.log(2)
    adjustment = np.maximum(-2.0, np.minimum(adjustment, 2.0))
    difficulty_bits = np.maximum(zero_bits + adjustment, 32.0)

    return difficulty_bits


def extract_design_matrix(desc, df):

    matrices = []

    for t in desc:

        df2 = df[list(t[1])]
        df2 = {'curr': df2.iloc[1:], 'prev': df2.iloc[:-1]}[t[0]]
        if t[2] == 'ewma':
            matrix = [(df2 if (alpha >= 1.0) else df2.ewm(alpha=alpha).mean()) for alpha in t[3]]
            matrix = [m.values for m in matrix]
        elif t[2] == 'fourier':
            matrix = df2.values * (2.0 * np.pi)
            matrix = [matrix * freq for freq in t[3]]
            matrix = [f(m) for m in matrix for f in [np.cos, np.sin]]
        else:
            raise ValueError("Unknown transformation: %s" % t[2])

        matrix = [m[:, np.newaxis, :] for m in matrix]
        matrix = np.concatenate(matrix, axis=1)

        matrices.append(matrix)

    return matrices


def update_design_matrix(desc, prev_matrices, prev_dict, curr_dict):

    matrices = []

    for (idx, t) in enumerate(desc):

        d = {'curr': curr_dict, 'prev': prev_dict}[t[0]]
        matrix = np.concatenate([d[a].reshape((-1, 1, 1)) for a in t[1]], axis=2)
        alphas = t[3].reshape((1, -1, 1))
        matrix = matrix * alphas

        if t[2] == 'ewma':
            matrix += prev_matrices[idx] * (1.0 - alphas)
        elif t[2] == 'fourier':
            matrix *= (2.0 * np.pi)
            matrix = np.concatenate([np.cos(matrix), np.sin(matrix)], axis=1)
        else:
            raise ValueError("Unknown transformation: %s" % t[2])

        matrices.append(matrix)

    return matrices


class Rollouts(object):

    def __init__(self, laststate, n_rollouts, models):

        self.laststate = laststate
        self.current_price = float(laststate['features']['price'])
        self.current_hashrate = float(laststate['features']['hashrate'])
        self.n_rollouts = n_rollouts
        self.gamma_params = np.array([48] * n_rollouts)
        self.models = models

        # Ensure all features and design matrices have shape (n_rollouts, ...)
        zero_array = np.array([0.0] * n_rollouts)

        f = self.laststate['features']
        for k in f:
            f[k] = f[k] + zero_array

        # Include height as features:
        f['start_height'] = self.laststate['phase'] * 48
        f['end_height']   = f['start_height'] + 48

        self.timeline = [f]

        for f in self.laststate['design_matrices'].values():
            for k in range(len(f)):
                z = zero_array.reshape((-1,) + (1,) * (len(f[k].shape) - 1))
                f[k] = f[k] + z

    def __getitem__(self, key):
        '''
        Return an array of shape (n_phases,) or (n_phases, n_rollouts)
        according to the specified key. Possible keys are:

        -- end_height : block height at the end of the phase;
        -- end_time : Unix time at the end of the phase;
        -- log_hashrate : natural logarithm of hashes per second;
        -- log_price : natural logarithm of BTC/USD exchange rate;
        -- log_reward : natural logarithm of total reward per block (in BTC);
        -- zero_bits : 32 + log2(difficulty).

        The first of these is 1-dimensional; all others are 2-dimensional.
        '''

        if key in self.timeline[-1]:
            return np.stack([t[key] for t in self.timeline], axis=0)
        elif ('log_' + key) in self.timeline[-1]:
            return np.exp(self['log_' + key])

    def run_model(self, modelname, currstate):

        dm = update_design_matrix(self.models[modelname]['predictors'],
                                  self.laststate['design_matrices'][modelname],
                                  self.laststate['features'],
                                  currstate['features'])
        currstate['design_matrices'][modelname] = dm
        X = np.concatenate([m.reshape((len(m), -1)) for m in dm], axis=1)
        sqrt_cov = self.models[modelname]['sqrt_cov']
        epsilon = np.dot(np.random.randn(self.n_rollouts, len(sqrt_cov)), sqrt_cov)
        Y = self.models[modelname]['model'].predict(X) + epsilon

        for (i, label) in enumerate(self.models[modelname]['responses']):
            currstate['features'][label] = Y[:, i]

    def iterate(self):

        phase = self.laststate['phase'] + 1
        height = phase * 48

        lastdict = self.laststate['features']

        zero_bits = lastdict['zero_bits']
        start_time = lastdict['end_time']
        last_retarget = lastdict['last_retarget']

        if (height % 2016) == 0:
            # difficulty adjustment
            zero_bits = get_new_difficulty_bits(zero_bits, start_time - last_retarget)
            last_retarget = start_time

        subsidy_btc = phase_to_subsidy(phase)

        if (height % 210000) == 0:
            # halving
            print("Block subsidy reduced to %.8f BTC at height %d." % (subsidy_btc, height))

        # The exact distribution of ilhr - log(hashrate). This is unbiased
        # (mean zero) but asymmetrically distributed.
        epsilon = eg48_mean - np.log(np.random.gamma(self.gamma_params, 1))

        currdict = {'start_height': height,
                    'end_height': height + 48,
                    'zero_bits': zero_bits,
                    'start_time': start_time,
                    'last_retarget': last_retarget,
                    'subsidy_btc': subsidy_btc,
                    'epsilon': epsilon}

        currstate = {'phase': phase, 'features': currdict, 'design_matrices': {}}

        self.run_model('fee', currstate)

        currdict['log_reward'] = np.log(np.exp(currdict['log_fees']) + subsidy_btc)

        self.run_model('hashrate', currstate)

        currdict['ilhr'] = currdict['log_hashrate'] + epsilon
        currdict['duration'] = np.exp(np.log(2) * currdict['zero_bits'] + eg48_mean - currdict['ilhr'])
        currdict['end_time'] = currdict['start_time'] + currdict['duration']

        self.laststate = currstate
        self.timeline.append(currdict)

    def resample_indices(self):

        et = self['end_time']
        ll = int(max(et[0])) // 86400
        ul = int(min(et[-1])) // 86400
        unix_times = np.array(range(ll, ul + 1)) * 86400

        as2 = lambda x : np.argsort(np.argsort(x, axis=0), axis=0)
        queries = unix_times.reshape((-1, 1)) + 0 * et[0]

        row_indices = as2(np.concatenate([queries, et], axis=0))[:len(queries)] - as2(queries)
        col_indices = np.array([list(range(len(et[0])))])

        return unix_times, row_indices, col_indices

    def to_timeline(self):

        unix_times, idxs, idxs2 = self.resample_indices()

        inv_cumsum = lambda x : np.concatenate([x[:1], np.diff(x, axis=0)], axis=0)

        # compute average block reward:
        block_reward = np.exp(self['log_reward'])
        c1 = inv_cumsum(np.cumsum(block_reward, axis=0)[idxs, idxs2])
        c2 = inv_cumsum(np.cumsum(block_reward / block_reward, axis=0)[idxs, idxs2])
        block_reward = c1 / (c2 + 1.0e-8)

        hashes_per_block = np.exp(np.log(2) * self['zero_bits'][idxs, idxs2])
        price = np.exp(self['log_price'][idxs, idxs2])

        return Timeline(unix_times, self.current_price, price, block_reward, hashes_per_block)



import numpy as np
import pandas as pd

from .history import *
from .rollouts import extract_design_matrix, Rollouts
from .parsing import parse_date, to_unix_time

from sklearn.linear_model import RidgeCV
from scipy.linalg import sqrtm


def halflives_to_alphas(halflives_in_days):

    halflives = np.maximum(1.0e-6, np.array(halflives_in_days) * 3.0)
    alphas = 1.0 - np.exp(np.log(0.5) / halflives)
    return alphas


class Simulator(object):

    def __init__(self, history=None, minphase=2520, maxphase=None, current_price=None):

        if history is None:
            history = download_history(current_price=current_price)

        self.history = history.copy()

        maxphase = maxphase or len(self.history)

        recency = np.maximum(maxphase - self.history['phase'].values, 1.0)
        weights = np.log(maxphase) - np.log(recency)
        weights[:minphase] *= 0.0
        weights[maxphase:] *= 0.0

        # truncate for technical reasons:
        self.weights = weights[1:]

        self.models = {}

    def plot_history(self, minx=None, maxx=None):

        if maxx is None:
            maxx = 48 * len(self.history)

        if minx is None:
            minx = max(0, maxx - 210000)

        from matplotlib import pyplot as plt

        if 'hashrate' not in self.models:
            self.train_hashrate_model()

        hashes_per_block = np.exp(self.history['zero_bits'] * np.log(2)).values
        block_heights = (self.history['phase'].values + 1) * 48

        df = pd.DataFrame({
            'empirical hashrate': self.history['hashrate'].values * 1.0e-18,
            'mining difficulty': hashes_per_block / 600e18,
            'price': np.nan
        }, index=block_heights)

        ax = df.loc[minx:maxx].plot(
            figsize=(12, 5),
            style=['y', 'b', 'r'],
            xlim=(minx, maxx),
            fontsize=14)

        plt.ylabel('global hashrate (EH/s)', fontsize=18)
        plt.xlabel('block height', fontsize=18)

        leg = ['empirical hashrate', 'mining difficulty', 'price', '3rd halving']
        for i in range(1, 34):
            hbh = 210000 * i
            if (minx <= hbh <= maxx):
                plt.axvline(hbh, color='grey', linestyle='--')
                ordinal_suffix = ('st' if i in [1, 21, 31] else ('nd' if i in [2, 22, 32] else ('rd' if i in [3, 23, 33] else 'th')));
                leg.append('%d%s halving' % (i, ordinal_suffix))
        plt.legend(leg, fontsize=16)

        df['price'] = self.history['price'].values
        df['price'].loc[minx:maxx].plot(ax=ax, style='r', secondary_y=True, alpha=0.4, fontsize=14)
        plt.ylabel('bitcoin price (USD)', fontsize=18)

        return ax


    def run_rollouts(self, n_rollouts, n_iters=0, until_height=None, until_time=None):

        if 'hashrate' not in self.models:
            self.train_hashrate_model()

        if 'fee' not in self.models:
            self.train_fee_model()

        laststate = {'phase': self.history['phase'].iloc[-1],
                    'features': {x : self.history[x].iloc[-1] for x in self.history.columns if x != 'phase'},
                    'design_matrices': {k : [x[-1:] for x in v['design_matrices']] for (k, v) in self.models.items()}}

        laststate['features']['last_retarget'] = self.history['start_time'].iloc[(laststate['phase'] // 42) * 42]

        r = Rollouts(laststate, n_rollouts, self.models)

        # if n_iters is specified, run that many iterations:
        for i in range(n_iters):
            r.iterate()

        # if until_height is specified, run until (at least) that height:
        if until_height is not None:
            while ((r.laststate['phase'] + 1) * 48 < until_height):
                r.iterate()

        # if until_time is specified, run until (at least) that time:
        if until_time is not None:

            until_time = to_unix_time(until_time)

            while (np.min(r.laststate['features']['end_time']) < until_time):
                r.iterate()

        return r

    def train_model(self, modelname, predictors, responses,
                    halflives=[0, 6, 24, 96, 384],
                    omegas=[1, 7],
                    model=RidgeCV, **kwargs):

        if len(halflives) > 0:
            # common usecase: autoregressive models use EWMAs of their own responses as predictors:
            predictors.append(('prev', tuple(responses), 'ewma', halflives_to_alphas(halflives)))

        if len(omegas) > 0:
            # common usecase: Fourier coefficients of time-of-week:
            predictors.append(('curr', ['start_time'], 'fourier', np.array(omegas) / (7 * 86400)))

        d = {}
        d['model'] = model(**kwargs)
        d['predictors'] = predictors
        d['responses'] = responses

        d['design_matrices'] = extract_design_matrix(predictors, self.history)

        d['X'] = np.concatenate([m.reshape((len(m), -1)) for m in d['design_matrices']], axis=1)
        d['Y'] = self.history[responses].iloc[1:].values
        d['model'].fit(d['X'], d['Y'], self.weights)

        # predictions:
        d['P'] = d['model'].predict(d['X'])

        # residuals:
        d['R'] = d['Y'] - d['P']

        # covariance matrix:
        d['cov'] = np.dot(d['R'].T, self.weights.reshape(-1, 1) * d['R']) / self.weights.sum()
        d['sqrt_cov'] = sqrtm(d['cov'])

        self.models[modelname] = d

    def train_hashrate_model(self, **kwargs):

        predictors = [('curr', ['log_reward', 'zero_bits'], 'ewma', halflives_to_alphas([0, 32]))]
        responses = ['ilhr', 'log_price']

        self.train_model('hashrate', predictors, responses, **kwargs)

        # remove unpredictable component:
        cov = self.models['hashrate']['cov']
        cov[0][0] -= (eg48_std ** 2)
        assert(np.linalg.det(cov) > 0.0)
        self.models['hashrate']['sqrt_cov'] = sqrtm(cov)

        # change prediction problem from ilhr to log_hashrate:
        self.models['hashrate']['responses'][0] = 'log_hashrate'

        # determine historical hashrate:
        epsilon_model = self.models['hashrate']['R'][:, 0]
        epsilon = np.concatenate([[0.0], epsilon_model])
        self.history['epsilon'] = epsilon
        self.history['log_hashrate'] = self.history['ilhr'] - self.history['epsilon']
        self.history['hashrate'] = np.exp(self.history['log_hashrate'])

    def train_fee_model(self, **kwargs):

        predictors = [('curr', ['epsilon'], 'ewma', halflives_to_alphas([0, 1, 4]))]
        responses = ['log_fees']

        self.train_model('fee', predictors, responses, **kwargs)




import re
import dateparser
from datetime import datetime, timezone

_dp_settings = {'PREFER_DATES_FROM': 'future',
                'TIMEZONE': 'UTC',
                'RELATIVE_BASE': datetime.now(tz=timezone.utc)}


def parse_date(input_date):

    if isinstance(input_date, str):

        input_date = input_date.strip()

        try:
            input_date = dateparser.parse(input_date, settings=_dp_settings)
        except:
            print("Fatal error occurred on date '%s'" % input_date)
            raise

    return input_date

# Checks that dateparser works correctly for distant future dates:
assert(parse_date('Mar 6, 2140').strftime('%Y-%m-%d') == '2140-03-06')


def to_unix_time(x):

    if isinstance(x, str):
        x = parse_date(x)

    if hasattr(x, 'timestamp'):
        x = x.timestamp()

    # round to nearest second:
    return int(x + 0.5)



import numpy as np
import pandas as pd

from .parsing import to_unix_time

class Timeline(object):

    def __init__(self, unix_times, current_price, price, block_reward, hashes_per_block):

        self.unix_times = unix_times
        self.pd_index = pd.to_datetime(unix_times, unit='s')
        self.current_price = current_price

        btc_per_th_day = 8.64e16 * block_reward / hashes_per_block

        # minimal sufficient statistics for mining profitability:
        self.rollouts = {'usd_per_btc': price, 'btc_per_th_day': btc_per_th_day}

    def evaluate_miner(self, miner_usd_cost, terahashes, daily_usd_cost, start_time, lifespan=5):

        start_time = to_unix_time(start_time)
        end_time = int(start_time + lifespan * 31556952)

        # the period of time for which the miner is online:
        active_mask = (self.unix_times >= start_time) & (self.unix_times <= end_time)
        active_th = active_mask.reshape((-1, 1)) * terahashes

        # compute gross and net income in both currencies:
        daily_btc_gross = active_th * self.rollouts['btc_per_th_day']
        daily_usd_gross = daily_btc_gross * self.rollouts['usd_per_btc']
        daily_usd_net = np.maximum(0.0, daily_usd_gross - daily_usd_cost)
        daily_btc_net = daily_usd_net / self.rollouts['usd_per_btc']

        miner_btc_cost = miner_usd_cost / self.current_price

        return {'daily_btc_gross': daily_btc_gross,
                'daily_btc_net': daily_btc_net,
                'miner_btc_cost': miner_btc_cost,
                'daily_usd_gross': daily_usd_gross,
                'daily_usd_net': daily_usd_net,
                'miner_usd_cost': miner_usd_cost}




import numpy as np
import pandas as pd

# Mean and standard deviation of the logarithm of the sum of 48
# independent identically distributed exponential random variables
# (known as ExpGammaDistribution[48,1,0] in Wolfram Mathematica).
eg48_mean = 3.8607481768292526
eg48_std = 0.14509257090677763

# We have log(duration) = log(2) * zero_bits + X - log(hashrate),
# where X is an ExpGammaDistribution[48,1,0] random variable and
# duration is the length of a phase (the interval between the
# (48n)th and (48(n+1))th blocks being mined).


def expand_history(df):
    '''
    Compute derived statistics from a history dataframe:
    '''

    df = df.copy()
    df['duration'] = df['end_time'] - df['start_time']
    df['log_fees'] = np.log(df['fees_btc'])
    df['log_reward'] = np.log(df['fees_btc'] + df['subsidy_btc'])
    df['log_price'] = np.log(df['price'])

    # Compute an unbiased (mean-zero) estimator of log(hashrate)
    # by reverse-engineering the formula for log(duration). This
    # estimator is called ilhr ('instantaneous log hashrate') and
    # has a standard deviation of eg48_std from true log(hashrate).
    ilhr = np.log(2) * df['zero_bits'] + eg48_mean - np.log(df['duration'])

    # The genesis block was mined at a much higher difficulty than
    # the remaining early blocks (40 zero_bits instead of 32) so
    # took 6 days to mine and severely skews the computation of
    # ilhr for the first phase. We overwrite it with an estimate
    # of Satoshi's initial hashrate from here:
    # http://organofcorti.blogspot.com/2014/08/167-satoshis-hashrate.html
    ilhr.iloc[:1] = 15.3

    df['ilhr'] = ilhr
    return df


def download_csv(csvname):

    print('Downloading CSV "%s" from Hatsya...' % csvname)
    df = pd.read_csv("https://catagolue.hatsya.com/textsamples/xs0_%s/b3s23/synthesis" % csvname)
    return df


def phase_to_subsidy(x):

    return 1.0e-8 * (5000000000 >> (x // 4375))


def download_history(current_price=None):

    df = download_csv('phases')

    subsidy_btc = df['phase'].apply(phase_to_subsidy)
    reward_btc = df['total_reward'] / 48e8

    # We ensure that fees are positive so that we can take the logarithm.
    # (Fees could be negative in practice if miners fail to claim the whole
    # block reward.) A lower bound of 1000 sats per block is conservative,
    # because the average number of transactions in a full block is >= 1000
    # and it is inconceivable that fees will drop below 1 sat per transaction
    # (that would equate to free transactions, since sats are indivisible!).
    fees_btc = np.maximum(0.00001000, reward_btc - subsidy_btc)

    price = 1e8 * df['total_reward_usd'] / df['total_reward']

    # Base early historical price on mining difficulty (NewLibertyStandard's
    # methodology). We use this up to block height 67200, after which we
    # switch to interpolated exchange data.
    difficulty = np.exp((df['zero_bits'] - 32) * np.log(2))
    for _ in range(2):
        difficulty = difficulty.rolling(42, min_periods=1).mean()
    price.iloc[:1400] = (difficulty / 1309.03).iloc[:1400]

    # Override last price:
    if current_price is not None:
        price.iloc[-1] = float(current_price)

    cols = {'phase': df['phase'],
            'start_time': df['start_time'],
            'end_time': df['end_time'],
            'zero_bits': df['zero_bits'],
            'fees_btc': fees_btc,
            'subsidy_btc': subsidy_btc,
            'price': price}

    df = pd.DataFrame(cols)
    assert np.all(df['phase'] == df.index)
    print('Downloaded history up to block height %d' % (48 * len(df)))

    # Include derived statistics:
    df = expand_history(df)

    return df



from setuptools import setup, find_packages

import hashboard
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), 'r') as f:
    long_description = f.read()

setup(
    name='hashboard',
    version=hashboard.__version__,
    description='Tool for estimating the profitability of bitcoin ASICs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Adam P. Goucher',
    author_email='adam@hatsya.com',
    url='https://gitlab.com/apgoucher/hashboard',
    license='MIT',
    packages=['hashboard'],
    # test_suite='lifelib.tests',
    include_package_data=True,
    zip_safe=False,
    install_requires=['numpy>=1.13', 'pandas>=1.3', 'scikit-learn>=1.0', 'seaborn>=0.11', 'scipy>=1.7', 'dateparser>=1.1'],
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',

        # MIT licence as always:
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.8',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
    ])
